#! /bin/bash

set -u
set -v

# Files and directories
binDir="bin"
calcDir="calculation"

# Param
binList="$(ls -1 "${binDir}"| sed '$!s/$/ /' | tr -d '\n')"
nbProdList="1" # Nombre de cores producteurs
typeProdList="none useless_loop line matrice" # Methode pour produire les valeurs
typeCacheList="L2 Memory" # Niveau de cache partage
perfOpt="stat -r 10 -e cycles -e L1-dcache-loads -e L1-dcache-stores -e L1-dcache-load-misses -e L1-dcache-store-misses -e L1-dcache-prefetch-misses"

# Const
nbIter="50000000" # Nb de lignes produites
sizeBuf="1" # En nombre de lignes de cache

# Nom generique des fichiers de log
logFileName="\$perfDirName/cache_\$typeCache-nbProd_\$nbProd-typeProd_\$typeProd-argTypeProd_\$argTypeProd-nbIter_\$nbIter-\$bin.log"
expDirName="logs"
perfDirName="$expDirName/perfCommMulti-`date +'%F-%Hh%Mm%S'`"

export LD_LIBRARY_PATH

[ -d "$expDirName" ] || mkdir "$expDirName"
[ -d "$perfDirName" ] || mkdir "$perfDirName"

function_run () {
	case $typeProd in
		"none" ) optTypeProd="" ;;
		"line" ) optTypeProd="-c lib/${calcDir}/libcalc_line.so ${argTypeProd}" ;;
		"matrice" ) optTypeProd="-c lib/${calcDir}/libcalc_mat.so ${argTypeProd}" ;;
		"useless_loop" ) optTypeProd="-c lib/${calcDir}/libcalc_useless_loop.so ${argTypeProd}" ;;
		* ) exit 1 ;;
	esac
	case $typeCache in
		"Memory" ) optTypeCache="" ;;
		"L2" ) optTypeCache="-s" ;;
		* ) exit 1 ;;
	esac

	make $binDir/$bin
	echo "On lance : \"perf $perfOpt $binDir/$bin $optTypeCache $optTypeProd -n $nbIter\""
	beginingDate=`date +%s`

	( perf $perfOpt $binDir/$bin $optTypeCache $optTypeProd -n $nbIter 2>&1 || echo "echec experience" ) | eval tee $logFileName
	endDate=`date +%s`
	duration_sec=`expr \( $endDate - $beginingDate \) % 60`
	duration_min=`expr \( $endDate - $beginingDate \) / 60`
	duration_hrs=`expr $duration_min / 60`
	duration_min=`expr $duration_min % 60`
	
	echo "Fin de l'experience : `date +'%H:%M:%S (%d/%m/%y)'`"
	echo "La duree est de : ${duration_hrs}hrs ${duration_min}min ${duration_sec}sec"
	echo ""
}



echo -e "On commence les perfs\n"
globalBeginingDate=`date +%s`

for nbProd in $nbProdList ; do
	for typeProd in $typeProdList; do
		for typeCache in $typeCacheList ; do
			for bin in $binList ; do
				case $typeProd in
					"useless_loop" )
						for argTypeProd in 1 `seq 5 5 50` ; do
							function_run ;
						done ;;
					"line" )
						for argTypeProd in 1 2 4 8 16 32 64 ; do
							function_run ;
						done ;;
					"matrice" )
						argTypeProd=16 ;
						function_run ;;
					* )
						argTypeProd=1 ;
						function_run ;;
				esac
			done
		done
	done
done

globalEndDate=`date +%s`
globalDuration_sec=`expr \( $globalEndDate - $globalBeginingDate \) % 60`
globalDuration_min=`expr \( $globalEndDate - $globalBeginingDate \) / 60`
globalDuration_hrs=`expr $globalDuration_min / 60`
globalDuration_min=`expr $globalDuration_min % 60`

echo "Fin de la serie d'experience : `date +'%H:%M:%S (%d/%m/%y)'`"
echo "La duree totale est de : ${globalDuration_hrs}hrs ${globalDuration_min}min ${globalDuration_sec}sec"

